@@ -3,6 +3,7 @@
 * $Version: $
 *
 * Copyright (c) Tanel Tammet 2004,2005,2006,2007,2008,2009
+* Copyright (c) Priit Järv 2009,2010,2011,2013,2014
 *
 * Contact: tanel.tammet@gmail.com
 *
@@ -94,6 +95,9 @@ void* wg_get_next_record(void* db, void* record); ///< returns NULL when error o
 void* wg_get_first_raw_record(void* db);
 void* wg_get_next_raw_record(void* db, void* record);
 
+void *wg_get_first_parent(void* db, void *record);
+void *wg_get_next_parent(void* db, void* record, void *parent);
+
 /* -------- setting and fetching record field values --------- */
 
 wg_int wg_get_record_len(void* db, void* record); ///< returns negative int when error
@@ -659,5 +663,8 @@ gint wg_encode_external_data(void *db, void *extdb, gint encoded);
 gint wg_translate_hdroffset(void *db, void *exthdr, gint encoded);
 void *wg_get_rec_owner(void *db, void *rec);
 #endif
+#ifdef USE_RECPTR_BITMAP
+gint wg_recptr_check(void *db,void *ptr);
+#endif
 
 #endif /* DEFINED_DBDATA_H */
