@@ -2,7 +2,7 @@
 * $Id:  $
 * $Version: $
 *
-* Copyright (c) Enar Reilent 2009, Priit Jrv 2010,2011
+* Copyright (c) Enar Reilent 2009, Priit Jrv 2010,2011,2013,2014
 *
 * This file is part of WhiteDB
 *
@@ -390,6 +390,11 @@ static int db_rotate_ttree(void *db, gint index_id, struct wg_tnode *root, int o
     //for later grandparent fix
     r = ee;
 
+  } else {
+    /* catch an error case (can't really happen) */
+    show_index_error(db, "tree rotate called with invalid argument, "\
+      "index may have become corrupt");
+    return -1;
   }
 
   //fix grandparent - regardless of current 'overweight' case
@@ -1584,7 +1589,7 @@ static gint hash_extend_prefix(void *db, wg_index_header *hdr, char *prefix,
   }
   if(prefix) {
     memcpy(newprefix, prefix, prefixlen);
-    newprefix[prefixlen] = '\0';
+    newprefix[prefixlen] = '\0'; /* XXX: why? double-check this */
   }
 
   memcpy(newprefix + (newlen - fldlen), fldbytes, fldlen);
@@ -2211,7 +2216,6 @@ gint wg_create_multi_index(void *db, gint *columns, gint col_count, gint type,
   /* create the actual index */
   switch(hdr->type) {
     case WG_INDEX_TYPE_TTREE:
-    case WG_INDEX_TYPE_TTREE_JSON:
       create_ttree_index(db, index_id);
       break;
     case WG_INDEX_TYPE_HASH:
@@ -2219,6 +2223,8 @@ gint wg_create_multi_index(void *db, gint *columns, gint col_count, gint type,
       if(create_hash_index(db, index_id))
         return -1;
       break;
+    case WG_INDEX_TYPE_TTREE_JSON:
+      /* Return an error, until proper implementation exists */
     default:
       show_index_error(db, "Invalid index type");
       return -1;
@@ -2338,7 +2344,7 @@ gint wg_drop_index(void *db, gint index_id){
       break;
     case WG_INDEX_TYPE_HASH:
     case WG_INDEX_TYPE_HASH_JSON:
-      if(drop_hash_index(db, index_id));
+      if(drop_hash_index(db, index_id))
         return -1;
       break;
     default:
@@ -2671,10 +2677,11 @@ gint wg_index_add_field(void *db, void *rec, gint column) {
   gint *ilist;
   gcell *ilistelem;
   db_memsegment_header* dbh = dbmemsegh(db);
+  gint reclen = wg_get_record_len(db, rec);
 
 #ifdef CHECK
   /* XXX: if used from wg_set_field() only, this is redundant */
-  if(column > MAX_INDEXED_FIELDNR || column >= wg_get_record_len(db, rec))
+  if(column > MAX_INDEXED_FIELDNR || column >= reclen)
     return -1;
   if(is_special_record(rec))
     return -1;
@@ -2692,8 +2699,10 @@ gint wg_index_add_field(void *db, void *rec, gint column) {
     if(ilistelem->car) {
       wg_index_header *hdr = \
         (wg_index_header *) offsettoptr(db, ilistelem->car);
-      if(MATCH_TEMPLATE(db, hdr, rec)) {
-        INDEX_ADD_ROW(db, hdr, ilistelem->car, rec)
+      if(reclen > hdr->rec_field_index[hdr->fields - 1]) {
+        if(MATCH_TEMPLATE(db, hdr, rec)) {
+          INDEX_ADD_ROW(db, hdr, ilistelem->car, rec)
+        }
       }
     }
     ilist = &ilistelem->cdr;
@@ -2710,8 +2719,10 @@ gint wg_index_add_field(void *db, void *rec, gint column) {
     if(ilistelem->car) {
       wg_index_header *hdr = \
         (wg_index_header *) offsettoptr(db, ilistelem->car);
-      if(MATCH_TEMPLATE(db, hdr, rec)) {
-        INDEX_ADD_ROW(db, hdr, ilistelem->car, rec)
+      if(reclen > hdr->rec_field_index[hdr->fields - 1]) {
+        if(MATCH_TEMPLATE(db, hdr, rec)) {
+          INDEX_ADD_ROW(db, hdr, ilistelem->car, rec)
+        }
       }
     }
     ilist = &ilistelem->cdr;
@@ -2751,12 +2762,15 @@ gint wg_index_add_rec(void *db, void *rec) {
       if(ilistelem->car) {
         wg_index_header *hdr = \
           (wg_index_header *) offsettoptr(db, ilistelem->car);
-        if(hdr->rec_field_index[0] >= i) {
-          /* A little trick: we only update index if the
-           * first column in the column list matches. The reasoning
-           * behind this is that we only want to update each index
-           * once, for multi-column indexes we can rest assured that
-           * the work was already done.
+        if(hdr->rec_field_index[hdr->fields - 1] == i) {
+          /* Only add the record if we're at the last column
+           * of the index. This way we ensure that a.) a record
+           * is entered once into a multi-column index and b.) the
+           * record is long enough so that it qualifies for the
+           * multi-column index.
+           * For a single-column index, the indexed column is
+           * also the last column, therefore the above is valid,
+           * altough the check is unnecessary.
            */
           if(MATCH_TEMPLATE(db, hdr, rec)) {
             INDEX_ADD_ROW(db, hdr, ilistelem->car, rec)
@@ -2798,7 +2812,8 @@ gint wg_index_add_rec(void *db, void *rec) {
               firstmatch = j;
           }
         }
-        if(firstmatch==i) {
+        if(firstmatch==i &&\
+          reclen > hdr->rec_field_index[hdr->fields - 1]) {
           /* The record matches AND this is the first time we
            * see this index. Update it.
            */
@@ -2825,10 +2840,11 @@ gint wg_index_del_field(void *db, void *rec, gint column) {
   gint *ilist;
   gcell *ilistelem;
   db_memsegment_header* dbh = dbmemsegh(db);
+  gint reclen = wg_get_record_len(db, rec);
 
 #ifdef CHECK
   /* XXX: if used from wg_set_field() only, this is redundant */
-  if(column > MAX_INDEXED_FIELDNR || column >= wg_get_record_len(db, rec))
+  if(column > MAX_INDEXED_FIELDNR || column >= reclen)
     return -1;
   if(is_special_record(rec))
     return -1;
@@ -2848,8 +2864,10 @@ gint wg_index_del_field(void *db, void *rec, gint column) {
       wg_index_header *hdr = \
         (wg_index_header *) offsettoptr(db, ilistelem->car);
 
-      if(MATCH_TEMPLATE(db, hdr, rec)) {
-        INDEX_REMOVE_ROW(db, hdr, ilistelem->car, rec)
+      if(reclen > hdr->rec_field_index[hdr->fields - 1]) {
+        if(MATCH_TEMPLATE(db, hdr, rec)) {
+          INDEX_REMOVE_ROW(db, hdr, ilistelem->car, rec)
+        }
       }
     }
     ilist = &ilistelem->cdr;
@@ -2864,8 +2882,10 @@ gint wg_index_del_field(void *db, void *rec, gint column) {
       wg_index_header *hdr = \
         (wg_index_header *) offsettoptr(db, ilistelem->car);
 
-      if(MATCH_TEMPLATE(db, hdr, rec)) {
-        INDEX_REMOVE_ROW(db, hdr, ilistelem->car, rec)
+      if(reclen > hdr->rec_field_index[hdr->fields - 1]) {
+        if(MATCH_TEMPLATE(db, hdr, rec)) {
+          INDEX_REMOVE_ROW(db, hdr, ilistelem->car, rec)
+        }
       }
     }
     ilist = &ilistelem->cdr;
@@ -2904,11 +2924,9 @@ gint wg_index_del_rec(void *db, void *rec) {
       if(ilistelem->car) {
         wg_index_header *hdr = \
           (wg_index_header *) offsettoptr(db, ilistelem->car);
-        if(hdr->rec_field_index[0] >= i) {
-          /* Ignore second, third etc references to multi-column
-           * indexes. XXX: This only works if index table is scanned
-           * sequentially, from position 0. See also comment for
-           * wg_index_add_rec command.
+        if(hdr->rec_field_index[hdr->fields - 1] == i) {
+          /* Only update once per index. See also comment for
+           * wg_index_add_rec function.
            */
           if(MATCH_TEMPLATE(db, hdr, rec)) {
             INDEX_REMOVE_ROW(db, hdr, ilistelem->car, rec)
@@ -2946,7 +2964,8 @@ gint wg_index_del_rec(void *db, void *rec) {
               firstmatch = j;
           }
         }
-        if(firstmatch==i) {
+        if(firstmatch==i &&\
+          reclen > hdr->rec_field_index[hdr->fields - 1]) {
           /* The record matches AND this is the first time we
            * see this index. Update it.
            */
