@@ -2,7 +2,7 @@
 * $Id:  $
 * $Version: $
 *
-* Copyright (c) Priit Jrv 2013
+* Copyright (c) Priit Jrv 2013, 2014
 *
 * This file is part of WhiteDB
 *
@@ -72,6 +72,13 @@ extern "C" {
 #define MAX_DEPTH 99 /* no reason to limit */
 #endif
 
+/* Commenting this out allows parsing literal value in input, but
+ * the current code lacks the capability of representing them
+ * (what record should they be stored in?) so there would be
+ * no obvious benefit.
+ */
+#define CHECK_TOPLEVEL_STRUCTURE
+
 typedef enum { ARRAY, OBJECT } stack_entry_t;
 
 struct __stack_entry_elem {
@@ -95,6 +102,7 @@ typedef struct {
   int stack_ptr;
   void *db;
   int isparam;
+  int isdocument;
   void **document;
 } parser_context;
 
@@ -107,7 +115,7 @@ static int add_key(parser_context *ctx, char *key);
 static int add_literal(parser_context *ctx, gint val);
 
 static gint run_json_parser(void *db, char *buf,
-  yajl_callbacks *cb, int isparam, void **document);
+  yajl_callbacks *cb, int isparam, int isdocument, void **document);
 static int check_push_cb(void* cb_ctx);
 static int check_pop_cb(void* cb_ctx);
 static int array_begin_cb(void* cb_ctx);
@@ -120,8 +128,9 @@ static int object_key_cb(void* cb_ctx, const unsigned char * strval,
                            size_t strl);
 static int elem_string_cb(void* cb_ctx, const unsigned char * strval,
                            size_t strl);
-static int pretty_print_json(void *db, FILE *f, void *rec,
-  int indent, int comma, int newline);
+static void print_cb(void *cb_ctx, const char *str, size_t len);
+static int pretty_print_json(void *db, yajl_gen *g, void *rec);
+static int pretty_print_jsonval(void *db, yajl_gen *g, gint enc);
 
 static gint show_json_error(void *db, char *errmsg);
 static gint show_json_error_fn(void *db, char *errmsg, char *filename);
@@ -172,7 +181,7 @@ yajl_callbacks input_cb = {
 gint wg_parse_json_file(void *db, char *filename) {
   char *buf = NULL;
   FILE *f = NULL;
-  int count = 0, result = 0, bufsize = 0, depth = 0;
+  int count = 0, result = 0, bufsize = 0, depth = -1;
   yajl_handle hand = NULL;
 
   buf = malloc(WG_JSON_INPUT_CHUNK);
@@ -240,8 +249,17 @@ gint wg_parse_json_file(void *db, char *filename) {
     goto done;
   }
 
+#ifdef CHECK_TOPLEVEL_STRUCTURE
+  if(depth == -1) {
+    show_json_error(db, "Top-level array or object is required in JSON");
+    result = -1;
+    goto done;
+  }
+#endif
+
+
   buf[count] = '\0';
-  result = wg_parse_json_document(db, buf);
+  result = wg_parse_json_document(db, buf, NULL);
 
 done:
   if(buf) free(buf);
@@ -250,27 +268,101 @@ done:
   return result;
 }
 
+/* Validate JSON data in a string buffer.
+ * Does not insert data into the database, so this may be used
+ * as a first pass before calling the wg_parse_*() functions.
+ *
+ * returns 0 for success.
+ * returns -1 in case of a syntax error.
+ */
+gint wg_check_json(void *db, char *buf) {
+  int count = 0, result = 0, depth = -1;
+  char *iptr = buf;
+  yajl_handle hand = NULL;
+
+#ifdef CHECK
+  if(!buf)
+    return show_json_error(db, "Invalid input buffer");
+#endif
+
+  /* setup parser */
+  hand = yajl_alloc(&validate_cb, NULL, (void *) &depth);
+  yajl_config(hand, yajl_allow_comments, 1);
+
+  while((count = strnlen(iptr, WG_JSON_INPUT_CHUNK)) > 0) {
+    if(yajl_parse(hand, (unsigned char *) iptr, count) != yajl_status_ok) {
+      show_json_error(db, "JSON parsing failed");
+      result = -1;
+      goto done;
+    }
+    iptr += count;
+  }
+
+  if(yajl_complete_parse(hand) != yajl_status_ok) {
+    show_json_error(db, "JSON parsing failed");
+    result = -1;
+  }
+#ifdef CHECK_TOPLEVEL_STRUCTURE
+  else if(depth == -1) {
+    show_json_error(db, "Top-level array or object is required in JSON");
+    result = -1;
+  }
+#endif
+
+done:
+  if(hand) yajl_free(hand);
+  return result;
+}
+
 /* Parse a JSON buffer.
  * The data is inserted in database using the JSON schema.
+ * If parsing is successful, the pointer referred to by
+ * **document will point to the top-level record.
+ * If **document is NULL, the pointer is discarded.
  *
  * returns 0 for success.
  * returns -1 on non-fatal error.
  * returns -2 if database is left non-consistent due to an error.
  */
-gint wg_parse_json_document(void *db, char *buf) {
-  void *document = NULL; /* ignore */
-  return run_json_parser(db, buf, &input_cb, 0, &document);
+gint wg_parse_json_document(void *db, char *buf, void **document) {
+  void *rec = NULL;
+  gint retv = run_json_parser(db, buf, &input_cb, 0, 1, &rec);
+  if(document)
+    *document = rec;
+  return retv;
+}
+
+/* Parse a JSON buffer.
+ * Like wg_parse_json_document, except the top-level object or
+ * array is not marked as a document.
+ *
+ * returns 0 for success.
+ * returns -1 on non-fatal error.
+ * returns -2 if database is left non-consistent due to an error.
+ */
+gint wg_parse_json_fragment(void *db, char *buf, void **document) {
+  void *rec = NULL;
+  gint retv = run_json_parser(db, buf, &input_cb, 0, 0, &rec);
+  if(document)
+    *document = rec;
+  return retv;
 }
 
 /* Parse a JSON parameter(s).
  * The data is inserted in database as "special" records.
+ * It does not make sense to call this function with NULL as the
+ * third parameter, as that would imply data input semantics but
+ * the records generated here are speficially flagged *non-data*.
  *
  * returns 0 for success.
  * returns -1 on non-fatal error.
  * returns -2 if database is left non-consistent due to an error.
  */
 gint wg_parse_json_param(void *db, char *buf, void **document) {
-  return run_json_parser(db, buf, &input_cb, 1, document);
+  if(!document) {
+    return show_json_error(db, "wg_parse_json_param: arg 3 cannot be NULL");
+  }
+  return run_json_parser(db, buf, &input_cb, 1, 1, document);
 }
 
 /* Run JSON parser.
@@ -280,6 +372,9 @@ gint wg_parse_json_param(void *db, char *buf, void **document) {
  * if isparam is specified, the data will not be indexed nor returned
  * by wg_get_*_record() calls.
  *
+ * if isdocument is 0, the input will be treated as a fragment and
+ * not as a full document.
+ *
  * if the call is successful, *document contains a pointer to the
  * top-level record.
  *
@@ -288,7 +383,7 @@ gint wg_parse_json_param(void *db, char *buf, void **document) {
  * returns -2 if database is left non-consistent due to an error.
  */
 static gint run_json_parser(void *db, char *buf,
-  yajl_callbacks *cb, int isparam, void **document)
+  yajl_callbacks *cb, int isparam, int isdocument, void **document)
 {
   int count = 0, result = 0;
   yajl_handle hand = NULL;
@@ -300,6 +395,7 @@ static gint run_json_parser(void *db, char *buf,
   ctx.stack_ptr = -1;
   ctx.db = db;
   ctx.isparam = isparam;
+  ctx.isdocument = isdocument;
   ctx.document = document;
 
   /* setup parser */
@@ -328,6 +424,7 @@ done:
 static int check_push_cb(void* cb_ctx)
 {
   int *depth = (int *) cb_ctx;
+  if(*depth == -1) *depth = 0; /* hack: something was pushed */
   if(++(*depth) >= MAX_DEPTH) {
     return 0;
   }
@@ -380,9 +477,11 @@ static int pop(parser_context *ctx)
   }
 
   if(e->type == ARRAY) {
-    rec = wg_create_array(ctx->db, e->size, istoplevel, ctx->isparam);
+    rec = wg_create_array(ctx->db, e->size,
+      (istoplevel && ctx->isdocument), ctx->isparam);
   } else {
-    rec = wg_create_object(ctx->db, e->size, istoplevel, ctx->isparam);
+    rec = wg_create_object(ctx->db, e->size,
+      (istoplevel && ctx->isdocument), ctx->isparam);
   }
 
   /* add elements to the database */
@@ -616,10 +715,20 @@ static int elem_string_cb(void* cb_ctx, const unsigned char * strval,
   return res;
 }
 
+static void print_cb(void *cb_ctx, const char *str, size_t len)
+{
+  FILE *f = (FILE *) cb_ctx;
+  fwrite(str, len, 1, f);
+}
+
 /*
- * Print a JSON document into the given stream.
+ * Print a JSON document. If a callback is given, it
+ * should be of type (void) (void *, char *, size_t) where the first
+ * pointer will be cast to FILE * stream. Otherwise the document will
+ * be written to stdout.
  */
-void wg_print_json_document(void *db, FILE *f, void *document) {
+void wg_print_json_document(void *db, void *cb, void *cb_ctx, void *document) {
+  yajl_gen g;
   if(!is_schema_document(document)) {
     /* Paranoia check. This increases the probability we're dealing
      * with records belonging to a proper schema. Omitting this check
@@ -628,7 +737,15 @@ void wg_print_json_document(void *db, FILE *f, void *document) {
     show_json_error(db, "Given record is not a document");
     return;
   }
-  pretty_print_json(db, f, document, 0, 0, 1);
+  g = yajl_gen_alloc(NULL);
+  yajl_gen_config(g, yajl_gen_beautify, 1);
+  if(cb) {
+    yajl_gen_config(g, yajl_gen_print_callback, (yajl_print_t *) cb, cb_ctx);
+  } else {
+    yajl_gen_config(g, yajl_gen_print_callback, print_cb, (void *) stdout);
+  }
+  pretty_print_json(db, &g, document);
+  yajl_gen_free(g);
 }
 
 /*
@@ -636,14 +753,14 @@ void wg_print_json_document(void *db, FILE *f, void *document) {
  * Returns 0 on success
  * Returns -1 on error.
  */
-static int pretty_print_json(void *db, FILE *f, void *rec,
-  int indent, int comma, int newline)
+static int pretty_print_json(void *db, yajl_gen *g, void *rec)
 {
   if(is_schema_object(rec)) {
     gint i, reclen;
 
-    /*OUT_INDENT(indent, i, f);*/
-    fprintf(f, "%s{\n", (comma ? "," : ""));
+    if(yajl_gen_map_open(*g) != yajl_gen_status_ok) {
+      return show_json_error(db, "Formatter failure");
+    }
 
     reclen = wg_get_record_len(db, rec);
     for(i=0; i<reclen; i++) {
@@ -652,72 +769,100 @@ static int pretty_print_json(void *db, FILE *f, void *rec,
       if(wg_get_encoded_type(db, enc) != WG_RECORDTYPE) {
         return show_json_error(db, "Object had an element of invalid type");
       }
-      if(pretty_print_json(db, f, wg_decode_record(db, enc), indent+1, i, 1)) {
+      if(pretty_print_json(db, g, wg_decode_record(db, enc))) {
         return -1;
       }
     }
 
-    OUT_INDENT(indent, i, f);
-    fprintf(f, "}%s", (newline ? "\n" : ""));
+    if(yajl_gen_map_close(*g) != yajl_gen_status_ok) {
+      return show_json_error(db, "Formatter failure");
+    }
   }
   else if(is_schema_array(rec)) {
     gint i, reclen;
 
-    fprintf(f, "%s[", (comma ? "," : ""));
+    if(yajl_gen_array_open(*g) != yajl_gen_status_ok) {
+      return show_json_error(db, "Formatter failure");
+    }
 
     reclen = wg_get_record_len(db, rec);
     for(i=0; i<reclen; i++) {
-      gint enc, type;
+      gint enc;
       enc = wg_get_field(db, rec, i);
-      type = wg_get_encoded_type(db, enc);
-      if(type == WG_RECORDTYPE) {
-        if(pretty_print_json(db, f, wg_decode_record(db, enc),
-          indent, i, 0)) {
-          return -1;
-        }
-      } else if(type == WG_STRTYPE) {
-        fprintf(f, "%s\"%s\"", (i ? "," : ""), wg_decode_str(db, enc));
-      } else {
-        /* other literal value */
-        char buf[80];
-        wg_snprint_value(db, enc, buf, 79);
-        fprintf(f, "%s%s", (i ? "," : ""), buf);
+      if(pretty_print_jsonval(db, g, enc)) {
+        return -1;
       }
     }
 
-    fprintf(f, "]%s", (newline ? "\n" : ""));
+    if(yajl_gen_array_close(*g) != yajl_gen_status_ok) {
+      return show_json_error(db, "Formatter failure");
+    }
   }
   else {
     /* assume key-value pair */
-    gint i, key, value, valtype;
+    gint key, value;
     key = wg_get_field(db, rec, WG_SCHEMA_KEY_OFFSET);
     value = wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET);
 
     if(wg_get_encoded_type(db, key) != WG_STRTYPE) {
       return show_json_error(db, "Key is of invalid type");
     } else {
-      OUT_INDENT(indent, i, f);
-      fprintf(f, "%s\"%s\": ", (comma ? "," : ""), wg_decode_str(db, key));
-    }
-
-    valtype = wg_get_encoded_type(db, value);
-    if(valtype == WG_RECORDTYPE) {
-      if(pretty_print_json(db, f, wg_decode_record(db, value),
-        indent, 0, 1)) {
-        return -1;
+      int len = wg_decode_str_len(db, key);
+      char *buf = wg_decode_str(db, key);
+      if(buf) {
+        if(yajl_gen_string(*g, (unsigned char *) buf,
+          (size_t) len) != yajl_gen_status_ok) {
+          return show_json_error(db, "Formatter failure");
+        }
       }
-    } else if(valtype == WG_STRTYPE) {
-      fprintf(f, "\"%s\"\n", wg_decode_str(db, value));
-    } else {
-      /* other literal value */
-      char buf[80];
-      wg_snprint_value(db, value, buf, 79);
-      fprintf(f, "%s\n", buf);
+    }
+    if(pretty_print_jsonval(db, g, value)) {
+      return -1;
     }
   }
   return 0;
 }
 
+/*
+ * Print a JSON array element or object value.
+ * May be an array or object itself.
+ */
+static int pretty_print_jsonval(void *db, yajl_gen *g, gint enc)
+{
+  gint type = wg_get_encoded_type(db, enc);
+  if(type == WG_RECORDTYPE) {
+    if(pretty_print_json(db, g, wg_decode_record(db, enc))) {
+      return -1;
+    }
+  } else if(type == WG_STRTYPE) {
+    int len = wg_decode_str_len(db, enc);
+    char *buf = wg_decode_str(db, enc);
+    if(buf) {
+      if(yajl_gen_string(*g, (unsigned char *) buf,
+        (size_t) len) != yajl_gen_status_ok) {
+        return show_json_error(db, "Formatter failure");
+      }
+    }
+  } else {
+    /* other literal value */
+    size_t len;
+    char buf[80];
+    wg_snprint_value(db, enc, buf, 79);
+    len = strlen(buf);
+    if(type == WG_INTTYPE || type == WG_DOUBLETYPE ||\
+      type == WG_FIXPOINTTYPE) {
+      if(yajl_gen_number(*g, buf, len) != yajl_gen_status_ok) {
+        return show_json_error(db, "Formatter failure");
+      }
+    } else {
+      if(yajl_gen_string(*g, (unsigned char *) buf,
+        len) != yajl_gen_status_ok) {
+        return show_json_error(db, "Formatter failure");
+      }
+    }
+  }
+  return 0;
+}
 
 /* ------------ error handling ---------------- */
 
