@@ -3,7 +3,7 @@
 * $Version: $
 *
 * Copyright (c) Andri Rebane 2009
-* Copyright (c) Priit Jrv 2013
+* Copyright (c) Priit Jrv 2013,2014
 *
 * This file is part of WhiteDB
 *
@@ -70,7 +70,6 @@ extern "C" {
 
 #ifdef _WIN32
 #define snprintf(s, sz, f, ...) _snprintf_s(s, sz+1, sz, f, ## __VA_ARGS__)
-#define inline _inline
 #endif
 
 #ifndef _WIN32
@@ -207,7 +206,6 @@ static int backup_journal(void *db, char *journal_fn) {
   return err;
 }
 
-
 /** Open the journal file.
  *
  * In create mode, we also take care of the backup copy.
@@ -221,18 +219,13 @@ static int open_journal(void *db, int create) {
   mode_t savemask = 0;
 #endif
 
-#ifndef _WIN32
-  snprintf(journal_fn, WG_JOURNAL_FN_BUFSIZE, "%s.%td",
-    WG_JOURNAL_FILENAME, dbh->key);
-#else
-  snprintf(journal_fn, WG_JOURNAL_FN_BUFSIZE, "%s.%Id",
-    WG_JOURNAL_FILENAME, dbh->key);
-#endif
-
+  wg_journal_filename(db, journal_fn, WG_JOURNAL_FN_BUFSIZE);
   if(create) {
 #ifndef _WIN32
     struct stat tmp;
-    savemask = umask(WG_JOURNAL_UMASK);
+    db_handle_logdata *ld = \
+      (db_handle_logdata *) (((db_handle *) db)->logdata);
+    savemask = umask(ld->umask);
     addflags |= O_CREAT;
 #else
     struct _stat tmp;
@@ -274,7 +267,7 @@ abort:
  *  be at least 9 bytes.
  *  based on http://stackoverflow.com/a/2982965
  */
-static inline size_t enc_varint(unsigned char *buf, wg_uint val) {
+static size_t enc_varint(unsigned char *buf, wg_uint val) {
   buf[0] = (unsigned char)(val | 0x80);
   if(val >= (1 << 7)) {
     buf[1] = (unsigned char)((val >>  7) | 0x80);
@@ -339,7 +332,7 @@ static inline size_t enc_varint(unsigned char *buf, wg_uint val) {
  *  assumes we're using a read buffer - this is acceptable and
  *  probably preferable when doing the journal replay.
  */
-static inline size_t dec_varint(unsigned char *buf, wg_uint *val) {
+static size_t dec_varint(unsigned char *buf, wg_uint *val) {
   wg_uint tmp = buf[0] & 0x7f;
   if(buf[0] & 0x80) {
     tmp |= ((buf[1] & 0x7f) << 7);
@@ -404,7 +397,7 @@ static inline size_t dec_varint(unsigned char *buf, wg_uint *val) {
  *  returns 0 on success
  *  returns -1 on error
  */
-static inline int fget_varint(void *db, FILE *f, wg_uint *val) {
+static int fget_varint(void *db, FILE *f, wg_uint *val) {
   register int c;
   wg_uint tmp;
 
@@ -613,7 +606,7 @@ static gint recover_journal(void *db, FILE *f, void *table)
 {
   int c;
   gint length = 0, offset = 0, newoffset;
-  gint col = 0, enc = 0, newenc;
+  gint col = 0, enc = 0, newenc, meta = 0;
   void *rec;
 
   for(;;) {
@@ -673,6 +666,13 @@ static gint recover_journal(void *db, FILE *f, void *table)
           return show_log_error(db, "Failed to set field data");
         }
         break;
+      case WG_JOURNAL_ENTRY_META:
+        GET_LOG_VARINT(db, f, offset, -1)
+        GET_LOG_VARINT(db, f, meta, -1)
+        newoffset = translate_offset(db, table, offset);
+        rec = offsettoptr(db, newoffset);
+        *((gint *) rec + RECORD_META_POS) = meta;
+        break;
       default:
         return show_log_error(db, "Invalid log entry");
     }
@@ -681,6 +681,24 @@ static gint recover_journal(void *db, FILE *f, void *table)
 }
 #endif /* USE_DBLOG */
 
+/** Return the name of the current journal
+ *
+ */
+void wg_journal_filename(void *db, char *buf, size_t buflen) {
+#ifdef USE_DBLOG
+  db_memsegment_header* dbh = dbmemsegh(db);
+
+#ifndef _WIN32
+  snprintf(buf, buflen, "%s.%td", WG_JOURNAL_FILENAME, dbh->key);
+#else
+  snprintf(buf, buflen, "%s.%Id", WG_JOURNAL_FILENAME, dbh->key);
+#endif
+  buf[buflen-1] = '\0';
+#else
+  buf[0] = '\0';
+#endif
+}
+
 /** Set up the logging area in the database handle
  *  Normally called when opening the database connection.
  */
@@ -720,6 +738,26 @@ void wg_cleanup_handle_logdata(void *db) {
 #endif
 }
 
+/** Set journal file umask.
+ *  This needs to be done separately from initializing the logging
+ *  data in the handle, as the mask may be derived from the
+ *  permissions of the shared memory segment and this is not
+ *  guaranteed to exist during the handle initialization.
+ *  Returns the old mask.
+ */
+int wg_log_umask(void *db, int cmask) {
+  int prev = 0;
+#ifdef USE_DBLOG
+  db_handle_logdata *ld = \
+    (db_handle_logdata *) (((db_handle *) db)->logdata);
+  if(ld) {
+    prev = ld->umask;
+    ld->umask = cmask & 0777;
+  }
+#endif
+  return prev;
+}
+
 /** Activate logging
  *
  * When successful, does the following:
@@ -961,6 +999,7 @@ static gint write_log_buffer(void *db, void *buf, int buflen)
  * WG_JOURNAL_ENTRY_ENC - encode a value (data bytes, extdata if applicable)
  *   followed by a single varint field that contains the encoded value
  * WG_JOURNAL_ENTRY_SET - set a field value (record offset, column, encoded value)
+ * WG_JOURNAL_ENTRY_META - set the metadata of a record
  *
  * lengths, offsets and encoded values are stored as varints
  */
@@ -1116,6 +1155,25 @@ gint wg_log_set_field(void *db, void *rec, gint col, gint data)
 #endif /* USE_DBLOG */
 }
 
+/** Log setting metadata
+ *
+ *  We assume that dbh->logging.active flag is checked before calling this.
+ */
+gint wg_log_set_meta(void *db, void *rec, gint meta)
+{
+#ifdef USE_DBLOG
+  unsigned char buf[1 + 2*VARINT_SIZE], *optr;
+  buf[0] = WG_JOURNAL_ENTRY_META;
+  optr = &buf[1];
+  optr += enc_varint(optr, (wg_uint) ptrtooffset(db, rec));
+  optr += enc_varint(optr, (wg_uint) meta);
+  return write_log_buffer(db, (void *) buf, optr - buf);
+#else
+  return show_log_error(db, "Logging is disabled");
+#endif /* USE_DBLOG */
+}
+
+
 /* ------------ error handling ---------------- */
 
 static gint show_log_error(void *db, char *errmsg) {
