@@ -2,7 +2,7 @@
 * $Id:  $
 * $Version: $
 *
-* Copyright (c) Priit Jrv 2010,2011,2013
+* Copyright (c) Priit Jrv 2010,2011,2013,2014
 *
 * This file is part of WhiteDB
 *
@@ -42,6 +42,7 @@ extern "C" {
 #include "dbcompare.h"
 #include "dbmpool.h"
 #include "dbschema.h"
+#include "dbhash.h"
 
 /* T-tree based scoring */
 #define TTREE_SCORE_EQUAL 5
@@ -106,6 +107,15 @@ static gint append_resultset(void *db, query_result_set *set, gint offset);
 static gint fetch_resultset(void *db, query_result_set *set);
 static query_result_set *intersect_resultset(void *db,
   query_result_set *seta, query_result_set *setb);
+static gint check_and_merge_by_kv(void *db, void *rec,
+  wg_json_query_arg *arg, query_result_set *next_set);
+static gint check_and_merge_by_key(void *db, void *rec,
+  wg_json_query_arg *arg, query_result_set *next_set);
+static gint check_and_merge_recursively(void *db, void *rec,
+  wg_json_query_arg *arg, query_result_set *next_set, int depth);
+static gint prepare_json_arglist(void *db, wg_json_query_arg *arglist,
+  wg_json_query_arg **sorted_arglist, gint argc,
+  gint *index_id, gint *vindex_id, gint *kindex_id);
 
 static gint encode_query_param_unistr(void *db, char *data, gint type,
   char *extdata, int length);
@@ -687,6 +697,7 @@ static wg_query *internal_build_query(void *db, void *matchrec, gint reclen,
   query = (wg_query *) malloc(sizeof(wg_query));
   if(!query) {
     show_query_error(db, "Failed to allocate memory");
+    if(full_arglist) free(full_arglist);
     return NULL;
   }
 
@@ -1422,33 +1433,83 @@ static gint fetch_resultset(void *db, query_result_set *set) {
   return 0;
 }
 
+#define NESTEDLOOP 0
+#define HASHJOIN 1
+
 /*
  * Create an intersection of two result sets.
+ * Join strategy:
+ *   if the number of inner loops expected is low (i.e. the sets
+ *   are small), nested loop join is used. Otherwise, hash join
+ *   is used.
+ *
  * Returns a new result set (can be empty).
  * Returns NULL on error.
  */
 static query_result_set *intersect_resultset(void *db,
   query_result_set *seta, query_result_set *setb)
 {
-  gint offseta;
   query_result_set *intersection;
+  int strategy = HASHJOIN;
 
   if(!(intersection = create_resultset(db))) {
     return NULL;
   }
+  if(seta->res_count * setb->res_count < 200) {
+    strategy = NESTEDLOOP; /* don't bother with hash table */
+  }
 
-  rewind_resultset(db, seta);
-  while((offseta = fetch_resultset(db, seta))) {
-    gint offsetb;
+  if(strategy == HASHJOIN) {
+    void *hasht = NULL;
+    gint offset;
+
+    if(seta->res_count > setb->res_count) {
+      query_result_set *tmp = seta;
+      seta = setb;
+      setb = tmp;
+    }
+
+    if(!(hasht = wg_dhash_init(db, seta->res_count))) {
+      free_resultset(db, intersection);
+      return NULL;
+    }
+
+    rewind_resultset(db, seta);
+    while((offset = fetch_resultset(db, seta))) {
+      if(wg_dhash_addkey(db, hasht, offset)) {
+        free_resultset(db, intersection);
+        wg_dhash_free(db, hasht);
+        return NULL;
+      }
+    }
     rewind_resultset(db, setb);
-    while((offsetb = fetch_resultset(db, setb))) {
-      if(offseta == offsetb) {
-        gint err = append_resultset(db, intersection, offseta);
+    while((offset = fetch_resultset(db, setb))) {
+      if(wg_dhash_haskey(db, hasht, offset)) {
+        gint err = append_resultset(db, intersection, offset);
         if(err) {
           free_resultset(db, intersection);
+          wg_dhash_free(db, hasht);
           return NULL;
         }
-        break;
+      }
+    }
+    wg_dhash_free(db, hasht);
+  }
+  else { /* nested loop strategy */
+    gint offseta;
+    rewind_resultset(db, seta);
+    while((offseta = fetch_resultset(db, seta))) {
+      gint offsetb;
+      rewind_resultset(db, setb);
+      while((offsetb = fetch_resultset(db, setb))) {
+        if(offseta == offsetb) {
+          gint err = append_resultset(db, intersection, offseta);
+          if(err) {
+            free_resultset(db, intersection);
+            return NULL;
+          }
+          break;
+        }
       }
     }
   }
@@ -1457,6 +1518,10 @@ static query_result_set *intersect_resultset(void *db,
 
 /*
  * Create a result set that contains only unique rows.
+ * Uniqueness test uses similar strategy to the intersect function
+ * (hash table for membership test, but revert to nested loop if
+ * low number of elements).
+ *
  * Returns a new result set (can be empty).
  * Returns NULL on error.
  */
@@ -1464,27 +1529,56 @@ static query_result_set *unique_resultset(void *db, query_result_set *set)
 {
   gint offset;
   query_result_set *unique;
+  int strategy = HASHJOIN;
 
   if(!(unique = create_resultset(db))) {
     return NULL;
   }
+  if(set->res_count < 20) {
+    strategy = NESTEDLOOP; /* don't bother with hash table */
+  }
 
   rewind_resultset(db, set);
-  while((offset = fetch_resultset(db, set))) {
-    gint offsetu, found = 0;
-    rewind_resultset(db, unique);
-    while((offsetu = fetch_resultset(db, unique))) {
-      if(offset == offsetu) {
-        found = 1;
-        break;
+
+  if(strategy == HASHJOIN) {
+    void *hasht = NULL;
+    if(!(hasht = wg_dhash_init(db, set->res_count))) {
+      free_resultset(db, unique);
+      return NULL;
+    }
+
+    while((offset = fetch_resultset(db, set))) {
+      if(!wg_dhash_haskey(db, hasht, offset)) {
+        gint err = append_resultset(db, unique, offset);
+        if(!err) {
+          err = wg_dhash_addkey(db, hasht, offset);
+        }
+        if(err) {
+          free_resultset(db, unique);
+          wg_dhash_free(db, hasht);
+          return NULL;
+        }
       }
     }
-    if(!found) {
-      /* We're now at the end of the set and may append normally. */
-      gint err = append_resultset(db, unique, offset);
-      if(err) {
-        free_resultset(db, unique);
-        return NULL;
+    wg_dhash_free(db, hasht);
+  }
+  else { /* nested loop */
+    while((offset = fetch_resultset(db, set))) {
+      gint offsetu, found = 0;
+      rewind_resultset(db, unique);
+      while((offsetu = fetch_resultset(db, unique))) {
+        if(offset == offsetu) {
+          found = 1;
+          break;
+        }
+      }
+      if(!found) {
+        /* We're now at the end of the set and may append normally. */
+        gint err = append_resultset(db, unique, offset);
+        if(err) {
+          free_resultset(db, unique);
+          return NULL;
+        }
       }
     }
   }
@@ -1493,19 +1587,241 @@ static query_result_set *unique_resultset(void *db, query_result_set *set)
 
 /* ------------------- (JSON) document query -------------------*/
 
-#define ADD_DOC_TO_RESULTSET(db, ns, cr, doc, err) \
+/* Note the non-conventional return code values:
+ * -1 adding the document failed
+ * 1 adding the document succeeded
+ * (0 is reserved for using this macro in a recursive function
+ * to differentiate between matches and non-matches)
+ */
+#define ADD_DOC_TO_RESULTSET(db, rec, ns, rc) \
+  void *doc = wg_find_document(db, rec); \
   if(doc) { \
-    err = append_resultset(db, ns, ptrtooffset(db, doc)); \
+    if(!append_resultset(db, ns, ptrtooffset(db, doc))) \
+      rc = 1; \
+    else \
+      rc = -1; \
   } else { \
-    err = show_query_error(db, "Failed to retrieve the document"); \
-  } \
-  if(err) { \
+    rc = show_query_error(db, "Failed to retrieve the document"); \
+  }
+
+#define IF_ERR_CLEAN_UP(db, cr, ns, al, rc) \
+  if(rc < 0) { \
     free_resultset(db, ns); \
     if(cr) \
       free_resultset(db, cr); \
+    if(al) \
+      free(al); \
     return NULL; \
   }
 
+#define ARGLIST_CLEANUP(al) \
+  if(al) \
+    free(al);
+
+#define ADD_DOC_ARRAY_UNWRAP(db, rec, ns, rc, k, v) \
+  void *arec = wg_decode_record(db, k); \
+  if(is_schema_array(arec)) { \
+    gint areclen = wg_get_record_len(db, arec); \
+    int j; \
+    for(j=0; j<areclen; j++) { \
+      if(WG_COMPARE(db, wg_get_field(db, arec, j), v) == WG_EQUAL) { \
+        ADD_DOC_TO_RESULTSET(db, rec, ns, rc) \
+        break; \
+      } \
+    } \
+  }
+
+/*
+ * Check if a record matches a key-value pair given in a query
+ * clause. If the value is an array in the record, each member
+ * of the array is compared to the value in the clause.
+ * (this behaviour emulates the JSON hash index, but can be disabled
+ * by #undef-ing JSON_SCAN_UNWRAP_ARRAY).
+ *
+ * returns 1 if the record matches and is added to the resultset
+ * returns 0 if the record does not match
+ * returns -1 if the record matches, but adding fails
+ */
+static gint check_and_merge_by_kv(void *db, void *rec,
+  wg_json_query_arg *arg, query_result_set *next_set)
+{
+  gint rc = 0;
+  gint reclen = wg_get_record_len(db, rec);
+  if(reclen > WG_SCHEMA_VALUE_OFFSET) { /* XXX: assume key
+                                         * before value */
+#ifndef JSON_SCAN_UNWRAP_ARRAY
+    if(WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_KEY_OFFSET),
+      arg->key) == WG_EQUAL &&\
+      WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET),
+      arg->value) == WG_EQUAL)
+    {
+      ADD_DOC_TO_RESULTSET(db, rec, next_set, rc)
+    }
+#else
+    if(WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_KEY_OFFSET),
+      arg->key) == WG_EQUAL) {
+      gint k = wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET);
+
+      if(WG_COMPARE(db, k, arg->value) == WG_EQUAL) {
+        /* Direct match. */
+        ADD_DOC_TO_RESULTSET(db, rec, next_set, rc)
+      } else if(wg_get_encoded_type(db, k) == WG_RECORDTYPE) {
+        /* No direct match, but if it is a record AND an array,
+         * scan the array contents.
+         */
+        ADD_DOC_ARRAY_UNWRAP(db, rec, next_set, rc, k, arg->value)
+      }
+    }
+#endif
+  }
+  return rc;
+}
+
+/*
+ * Like check_and_merge_by_kv() except key comparison is skipped
+ * (i.e. the caller is iterating over key index)
+ */
+static gint check_and_merge_by_key(void *db, void *rec,
+  wg_json_query_arg *arg, query_result_set *next_set)
+{
+  gint rc = 0;
+  gint reclen = wg_get_record_len(db, rec);
+  if(reclen > WG_SCHEMA_VALUE_OFFSET) {
+#ifndef JSON_SCAN_UNWRAP_ARRAY
+    if(WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET),
+      arg->value) == WG_EQUAL)
+    {
+      ADD_DOC_TO_RESULTSET(db, rec, next_set, rc)
+    }
+#else
+    gint k = wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET);
+
+    if(WG_COMPARE(db, k, arg->value) == WG_EQUAL) {
+      ADD_DOC_TO_RESULTSET(db, rec, next_set, rc)
+    } else if(wg_get_encoded_type(db, k) == WG_RECORDTYPE) {
+      ADD_DOC_ARRAY_UNWRAP(db, rec, next_set, rc, k, arg->value)
+    }
+#endif
+  }
+  return rc;
+}
+
+/*
+ * Check if the record or any of its children matches
+ * the given key/value pair. The search is stopped upon
+ * the first match.
+ *
+ * returns 1 if the record matches and is added to the resultset
+ * returns 0 if the record does not match
+ * returns -1 if the record matches, but adding fails
+ */
+static gint check_and_merge_recursively(void *db, void *rec,
+  wg_json_query_arg *arg, query_result_set *next_set, int depth)
+{
+  gint i, reclen, rc;
+  rc = check_and_merge_by_kv(db, rec, arg, next_set);
+  if(rc) /* successful match or an error */
+    return rc;
+
+  if(depth <= 0) {
+    return show_query_error(db, "scanning document: recursion too deep");
+  }
+  reclen = wg_get_record_len(db, rec);
+  for(i=0; i<reclen; i++) {
+    gint enc = wg_get_field(db, rec, i);
+    gint type = wg_get_encoded_type(db, enc);
+    if(type == WG_RECORDTYPE) {
+      rc = check_and_merge_recursively(db, wg_decode_record(db, enc),
+        arg, next_set, depth-1);
+      if(rc) /* successful match or an error */
+        return rc;
+    }
+  }
+  return 0; /* no match */
+}
+
+/* Prepare argument list. This sorts clauses that are either less
+ * costly to query or restrict the following processing the most
+ * (not yet implemented, depends on statistics). Also determines
+ * which indexes can and should be used.
+ *
+ * Returns 0 on success.
+ * Returns -1 on error.
+ * in case of error, the contents of return parameters are unmodified.
+ * in case of success, **sorted_arglist may be set to NULL, if the
+ * argument list does not require sorting. The caller should always
+ * check that.
+ */
+static gint prepare_json_arglist(void *db, wg_json_query_arg *arglist,
+  wg_json_query_arg **sorted_arglist, gint argc,
+  gint *index_id, gint *vindex_id, gint *kindex_id)
+{
+  gint icols[2], need_ttree = 0;
+  wg_json_query_arg *tmp = NULL;
+
+  /* Get index */
+  icols[0] = WG_SCHEMA_KEY_OFFSET;
+  icols[1] = WG_SCHEMA_VALUE_OFFSET;
+  *index_id = wg_multi_column_to_index_id(db, icols, 2,
+    WG_INDEX_TYPE_HASH_JSON, NULL, 0);
+  *vindex_id = *kindex_id = -1;
+
+  if(argc > 1) {
+    /* There is something to sort. In the future we can also sort by
+     * cardinality here (provided that stats are available). */
+    gint i, j;
+    tmp = malloc(sizeof(wg_json_query_arg) * argc);
+    if(!tmp) {
+      return show_query_error(db, "Failed to prepare query arguments");
+    }
+
+    /* First pass: literal values only */
+    for(i=0, j=0; i<argc; i++) {
+      if(wg_get_encoded_type(db, arglist[i].value) != WG_RECORDTYPE) {
+        tmp[j].key = arglist[i].key;
+        tmp[j++].value = arglist[i].value;
+      }
+    }
+
+    /* Was there something left? In that case, we might need T-tree
+     * to speed up scanning for the remainder of clauses. We also use
+     * T-tree if hash is not available at all.
+     */
+    if(j<i) {
+      need_ttree = 1;
+    }
+
+    /* Second pass: complex structures only */
+    for(i=0; i<argc; i++) {
+      if(wg_get_encoded_type(db, arglist[i].value) == WG_RECORDTYPE) {
+        tmp[j].key = arglist[i].key;
+        tmp[j++].value = arglist[i].value;
+      }
+    }
+  } else {
+    /* Complex structures are not present in the hash index */
+    if(wg_get_encoded_type(db, arglist[0].value) == WG_RECORDTYPE) {
+      need_ttree = 1;
+    }
+  }
+
+  /* Get T-tree index if needed. Value index is preferred, but
+   * it must be of the type that supports array unwrap. Otherwise
+   * we'll settle for a key index.
+   */
+  if(*index_id == -1 || need_ttree) {
+    *vindex_id = wg_multi_column_to_index_id(db, &icols[1], 1,
+      WG_INDEX_TYPE_TTREE_JSON, NULL, 0);
+    if(*vindex_id == -1) {
+      *kindex_id = wg_multi_column_to_index_id(db, &icols[0], 1,
+        WG_INDEX_TYPE_TTREE, NULL, 0);
+    }
+  }
+
+  *sorted_arglist = tmp;
+  return 0;
+}
+
 /*
  * Find a list of documents that contain the key-value pairs.
  * Returns a prefetch query object.
@@ -1514,8 +1830,9 @@ static query_result_set *unique_resultset(void *db, query_result_set *set)
 wg_query *wg_make_json_query(void *db, wg_json_query_arg *arglist, gint argc) {
   wg_query *query = NULL;
   query_result_set *curr_res = NULL;
-  gint index_id = -1;
-  gint icols[2], i;
+  wg_json_query_arg *sorted_arglist = NULL;
+  gint index_id = -1, vindex_id = -1, kindex_id = -1;
+  gint i;
 
 #ifdef CHECK
   if(!arglist || argc < 1) {
@@ -1531,18 +1848,21 @@ wg_query *wg_make_json_query(void *db, wg_json_query_arg *arglist, gint argc) {
   }
 #endif
 
-  /* Get index */
-  icols[0] = WG_SCHEMA_KEY_OFFSET;
-  icols[1] = WG_SCHEMA_VALUE_OFFSET;
-  index_id = wg_multi_column_to_index_id(db, icols, 2,
-    WG_INDEX_TYPE_HASH_JSON, NULL, 0);
+  /* Sort the argument list. This also checks for usable indexes, so
+   * we're calling it even if we have just one argument.
+   */
+  prepare_json_arglist(db, arglist, &sorted_arglist, argc,
+    &index_id, &vindex_id, &kindex_id);
+  /* HACK: this way, the following code does not need to care
+   * whether we sorted the argument list or not.
+   */
+  if(sorted_arglist)
+    arglist = sorted_arglist;
 
   /* Iterate over the argument pairs.
    * XXX: it is possible that getting the first set from index and
    * doing a scan to check the remaining arguments is faster than
    * doing the intersect operation of sets retrieved from index.
-   * XXX: given that we don't index complex structures, reorder
-   * arguments so that immediate values come first.
    */
   for(i=0; i<argc; i++) {
     query_result_set *next_set, *tmp_set;
@@ -1571,65 +1891,88 @@ wg_query *wg_make_json_query(void *db, wg_json_query_arg *arglist, gint argc) {
         gint *nextoffset = &reclist_offset;
         while(*nextoffset) {
           gcell *rec_cell = (gcell *) offsettoptr(db, *nextoffset);
-          gint err = -1;
-          void *document = \
-            wg_find_document(db, offsettoptr(db, rec_cell->car));
-          ADD_DOC_TO_RESULTSET(db, next_set, curr_res, document, err)
+          gint rc = -1;
+          ADD_DOC_TO_RESULTSET(db, offsettoptr(db, rec_cell->car),
+            next_set, rc)
+          IF_ERR_CLEAN_UP(db, curr_res, next_set, sorted_arglist, rc)
           nextoffset = &(rec_cell->cdr);
         }
       }
     }
+#if 0
+    else if(vindex_id > 0) {
+      /* XXX: unimplemented: scan T-tree for values */
+    }
+#endif
+    else if(kindex_id > 0) {
+      /* Hash index not usable, do a scan but leverage an index on the
+       * key field to reduce the number of records visited.
+       */
+      gint curr_offset = 0, curr_slot = -1, end_offset = 0, end_slot = -1;
+
+      if(find_ttree_bounds(db, kindex_id, WG_SCHEMA_KEY_OFFSET,
+          arglist[i].key, arglist[i].key, 1, 1,
+          &curr_offset, &curr_slot, &end_offset, &end_slot)) {
+        curr_offset = 0;
+      }
+
+      while(curr_offset) {
+        gint rc;
+        struct wg_tnode *node = (struct wg_tnode *) offsettoptr(db, curr_offset);
+        void *rec = offsettoptr(db, node->array_of_values[curr_slot]);
+
+        rc = check_and_merge_by_key(db, rec, &arglist[i], next_set);
+        IF_ERR_CLEAN_UP(db, curr_res, next_set, sorted_arglist, rc)
+
+        if(curr_offset==end_offset && curr_slot==end_slot) {
+          break;
+        } else {
+          curr_slot += 1; /* direction implied as 1 */
+          if(curr_slot >= node->number_of_elements) {
+#ifdef CHECK
+            if(end_offset==curr_offset) {
+              show_query_error(db, "Warning: end slot mismatch, possible bug");
+              break;
+            } else {
+#endif
+              curr_offset = TNODE_SUCCESSOR(db, node);
+              curr_slot = 0;
+#ifdef CHECK
+            }
+#endif
+          }
+        }
+      }
+    }
+    else if(curr_res) {
+      /* No index, do a scan over the current resultset. This also happens if
+       * the value is a complex structure.
+       */
+      gint offset;
+      rewind_resultset(db, curr_res);
+      while((offset = fetch_resultset(db, curr_res))) {
+        gint *rec = offsettoptr(db, offset);
+#ifndef USE_BACKLINKING
+        gint rc = check_and_merge_recursively(db,
+          rec, &arglist[i], next_set, 99);
+#else
+        gint rc = check_and_merge_recursively(db,
+          rec, &arglist[i], next_set, WG_COMPARE_REC_DEPTH);
+#endif
+        IF_ERR_CLEAN_UP(db, curr_res, next_set, sorted_arglist, rc)
+      }
+      /* Skip merge in this iteration, next_set is a subset of curr_res */
+      free_resultset(db, curr_res);
+      curr_res = NULL;
+    }
     else {
-      /* No index, do a scan. This also happens if the value
-       * is a complex structure.
-       * XXX: if i>0 scan curr_res instead! (duh) */
+      /* No index and no intermediate result to use, full
+       * scan of database required.
+       */
       gint *rec = wg_get_first_record(db);
       while(rec) {
-        gint reclen = wg_get_record_len(db, rec);
-        if(reclen > WG_SCHEMA_VALUE_OFFSET) { /* XXX: assume key
-                                               * before value */
-#ifndef JSON_SCAN_UNWRAP_ARRAY
-          if(WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_KEY_OFFSET),
-            arglist[i].key) == WG_EQUAL &&\
-            WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET),
-            arglist[i].value) == WG_EQUAL)
-          {
-            gint err = -1;
-            void *document = wg_find_document(db, rec);
-            ADD_DOC_TO_RESULTSET(db, next_set, curr_res, document, err)
-          }
-#else
-          if(WG_COMPARE(db, wg_get_field(db, rec, WG_SCHEMA_KEY_OFFSET),
-            arglist[i].key) == WG_EQUAL) {
-            gint k = wg_get_field(db, rec, WG_SCHEMA_VALUE_OFFSET);
-
-            if(WG_COMPARE(db, k, arglist[i].value) == WG_EQUAL) {
-              /* Direct match. */
-              gint err = -1;
-              void *document = wg_find_document(db, rec);
-              ADD_DOC_TO_RESULTSET(db, next_set, curr_res, document, err)
-            } else if(wg_get_encoded_type(db, k) == WG_RECORDTYPE) {
-              /* No direct match, but if it is a record AND an array,
-               * scan the array contents.
-               */
-              void *arec = wg_decode_record(db, k);
-              if(is_schema_array(arec)) {
-                gint areclen = wg_get_record_len(db, arec);
-                int j;
-                for(j=0; j<areclen; j++) {
-                  if(WG_COMPARE(db, wg_get_field(db, arec, j),
-                   arglist[i].value) == WG_EQUAL) {
-                    gint err = -1;
-                    void *document = wg_find_document(db, rec);
-                    ADD_DOC_TO_RESULTSET(db, next_set, curr_res, document, err)
-                    break;
-                  }
-                }
-              }
-            }
-          }
-#endif
-        }
+        gint rc = check_and_merge_by_kv(db, rec, &arglist[i], next_set);
+        IF_ERR_CLEAN_UP(db, curr_res, next_set, sorted_arglist, rc)
         rec = wg_get_next_record(db, rec);
       }
     }
@@ -1640,22 +1983,20 @@ wg_query *wg_make_json_query(void *db, wg_json_query_arg *arglist, gint argc) {
     if(!tmp_set) {
       if(curr_res)
         free_resultset(db, curr_res);
+      ARGLIST_CLEANUP(sorted_arglist)
       return NULL;
     } else {
       next_set = tmp_set;
     }
 
     /* Update the query result */
-    if(i) {
+    if(curr_res) {
       /* Working resultset exists, create an intersection */
-      if(curr_res->res_count < next_set->res_count) { /* minor optimization */
-        tmp_set = intersect_resultset(db, curr_res, next_set);
-      } else {
-        tmp_set = intersect_resultset(db, next_set, curr_res);
-      }
+      tmp_set = intersect_resultset(db, curr_res, next_set);
       free_resultset(db, curr_res);
       free_resultset(db, next_set);
       if(!tmp_set) {
+        ARGLIST_CLEANUP(sorted_arglist)
         return NULL;
       } else {
         curr_res = tmp_set;
@@ -1665,6 +2006,7 @@ wg_query *wg_make_json_query(void *db, wg_json_query_arg *arglist, gint argc) {
       curr_res = next_set;
     }
   }
+  ARGLIST_CLEANUP(sorted_arglist)
 
   /* Initialize query object */
   query = (wg_query *) malloc(sizeof(wg_query));
