@@ -3,12 +3,11 @@
 * $Version: $
 *
 * Copyright (c) Tanel Tammet 2004,2005,2006,2007,2008,2009
+* Copyright (c) Enar Reilent 2009
+* Copyright (c) Priit Jrv 2010,2011,2012,2013,2014
 *
 * Contact: tanel.tammet@gmail.com
 *
-* Command parser written by Priit Jrv, some commands written
-* by Enar Reilent.
-*
 * This file is part of WhiteDB
 *
 * WhiteDB is free software: you can redistribute it and/or modify
@@ -39,6 +38,12 @@
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
+#ifndef _WIN32
+#include <sys/types.h>
+#include <pwd.h>
+#include <grp.h>
+#endif
+
 #ifdef _WIN32
 #include <conio.h> // for _getch
 #endif
@@ -55,7 +60,6 @@ extern "C" {
 #include "../Db/dballoc.h"
 #include "../Db/dbmem.h"
 #include "../Db/dbdata.h"
-#include "../Db/dbtest.h"
 #include "../Db/dbdump.h"
 #include "../Db/dblog.h"
 #include "../Db/dbquery.h"
@@ -74,7 +78,6 @@ extern "C" {
 #define sscanf sscanf_s  /* XXX: This will break for string parameters */
 #endif
 
-#define TESTREC_SIZE 3
 #define FLAGS_FORCE 0x1
 #define FLAGS_LOGGING 0x2
 
@@ -106,6 +109,7 @@ extern "C" {
 
 gint parse_shmsize(char *arg);
 gint parse_flag(char *arg);
+int parse_memmode(char *arg);
 wg_query_arg *make_arglist(void *db, char **argv, int argc, int *sz);
 void free_arglist(void *db, wg_query_arg *arglist, int sz);
 void query(void *db, char **argv, int argc);
@@ -115,6 +119,8 @@ int add_row(void *db, char **argv, int argc);
 wg_json_query_arg *make_json_arglist(void *db, char *json, int *sz,
  void **doc);
 void findjson(void *db, char *json);
+void segment_stats(void *db);
+void print_indexes(void *db, FILE *f);
 
 
 /* ====== Functions ============== */
@@ -159,25 +165,27 @@ void usage(char *prog) {
 #ifdef USE_DBLOG
   printf("    replay <filename> - replay a journal file.\n");
 #endif
-  printf("    test - run quick database tests.\n"\
-    "    fulltest - run in-depth database tests.\n"\
-    "    header - print header data.\n"\
-    "    fill <nr of rows> [asc | desc | mix] - fill db with integer data.\n"\
+  printf("    info - print information about the memory database.\n"\
     "    add <value1> .. - store data row (only int or str recognized)\n"\
     "    select <number of rows> [start from] - print db contents.\n"\
     "    query <col> \"<cond>\" <value> .. - basic query.\n"\
     "    del <col> \"<cond>\" <value> .. - like query. Matching rows "\
     "are deleted from database.\n"\
-    "    addjson [<filename>] - store a json document.\n"\
-    "    findjson <json> - find documents with matching keys/values.\n");
+    "    addjson [filename] - store a json document.\n"\
+    "    findjson <json> - find documents with matching keys/values.\n"\
+    "    createindex <column> - create ttree index\n" \
+    "    createhash <columns> - create hash index (JSON support)\n" \
+    "    dropindex <index id> - delete an index\n" \
+    "    listindex - list all indexes in database\n");
 #ifdef _WIN32
   printf("    server [-l] [size] - provide persistent shared memory for "\
     "other processes (-l: enable logging in the database). Will allocate "\
     "requested amount of memory and sleep; "\
     "Ctrl+C aborts and releases the memory.\n");
 #else
-  printf("    create [-l] [size] - create empty db of given size "\
-    "(-l: enable logging in the database).\n");
+  printf("    create [-l] [size [mode]] - create empty db of given size "\
+    "(-l: enable logging in the database, mode: segment permissions "\
+    "(octal)).\n");
 #endif
   printf("\nCommands may have variable number of arguments. "\
     "Commands that take values as arguments have limited support "\
@@ -245,6 +253,19 @@ gint parse_flag(char *arg) {
   return 0;
 }
 
+/** Parse the mode bits given in octal (textual representation)
+ *
+ */
+int parse_memmode(char *arg) {
+  char *trailing = NULL;
+  long parsed = strtol(arg, &trailing, 8);
+  if(errno == 0 && (!trailing || strlen(trailing) == 0) &&
+      parsed <= 0777 && parsed > 0) {
+    return (int) parsed;
+  }
+  return 0;
+}
+
 /** top level for the database command line tool
 *
 *
@@ -303,7 +324,7 @@ int main(int argc, char **argv) {
       }
 
       shmptr=wg_attach_memsegment(shmname, minsize, maxsize, 1,
-        (flags & FLAGS_LOGGING));
+        (flags & FLAGS_LOGGING), 0);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -333,7 +354,7 @@ int main(int argc, char **argv) {
         }
       }
 
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -376,15 +397,15 @@ int main(int argc, char **argv) {
     }
 #endif
     else if(argc>(i+1) && !strcmp(argv[i],"exportcsv")){
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
       }
 
-      RLOCK(shmptr, wlock);
+      RLOCK(shmptr, rlock);
       wg_export_db_csv(shmptr,argv[i+1]);
-      RULOCK(shmptr, wlock);
+      RULOCK(shmptr, rlock);
       break;
     }
     else if(argc>(i+1) && !strcmp(argv[i],"importcsv")){
@@ -449,7 +470,7 @@ int main(int argc, char **argv) {
     else if(argc>i && !strcmp(argv[i],"runreasoner")){
       wg_int err;
 
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -463,17 +484,6 @@ int main(int argc, char **argv) {
       //break;
       break;
     }
-    else if(argc>i && !strcmp(argv[i],"testreasoner")){
-      wg_int err;
-      //printf("about to call wg_test_reasoner\n");
-      err = wg_test_reasoner(argc,argv);
-      //if(!err);
-        //printf("wg_test_reasoner finished ok.\n");
-      //else
-        //fprintf(stderr, "wg_test_reasoner finished with an error %d.\n",err);
-      //break;
-      break;
-    }
 
 #endif
 
@@ -482,16 +492,16 @@ int main(int argc, char **argv) {
       wg_int err;
       int pref_fields = atol(argv[i+1]);
 
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
       }
 
       printf("Exporting with %d prefix fields.\n", pref_fields);
-      RLOCK(shmptr, wlock);
+      RLOCK(shmptr, rlock);
       err = wg_export_raptor_rdfxml_file(shmptr, pref_fields, argv[i+2]);
-      RULOCK(shmptr, wlock);
+      RULOCK(shmptr, rlock);
       if(err)
         fprintf(stderr, "Export failed.\n");
       break;
@@ -523,24 +533,15 @@ int main(int argc, char **argv) {
       break;
     }
 #endif
-    else if(!strcmp(argv[i],"test")) {
-      /* This test function does it's own memory allocation. */
-      wg_run_tests(WG_TEST_QUICK, 2);
-      break;
-    }
-    else if(!strcmp(argv[i],"fulltest")) {
-      wg_run_tests(WG_TEST_FULL, 2);
-      break;
-    }
-    else if(!strcmp(argv[i], "header")) {
-      shmptr=wg_attach_database(shmname, shmsize);
+    else if(!strcmp(argv[i], "info")) {
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
       }
-      RLOCK(shmptr, wlock);
-      wg_show_db_memsegment_header(shmptr);
-      RULOCK(shmptr, wlock);
+      RLOCK(shmptr, rlock);
+      segment_stats(shmptr);
+      RULOCK(shmptr, rlock);
       break;
     }
 #ifdef _WIN32
@@ -556,7 +557,7 @@ int main(int argc, char **argv) {
           fprintf(stderr, "Failed to parse memory size, using default.\n");
       }
       shmptr=wg_attach_memsegment(shmname, shmsize, shmsize, 1,
-        (flags & FLAGS_LOGGING));
+        (flags & FLAGS_LOGGING), 0);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -567,7 +568,7 @@ int main(int argc, char **argv) {
     }
 #else
     else if(!strcmp(argv[i],"create")) {
-      int flags = 0;
+      int flags = 0, mode = 0;
       if(argc>(i+1) && argv[i+1][0] == '-') {
         flags = parse_flag(argv[++i]);
       }
@@ -577,8 +578,13 @@ int main(int argc, char **argv) {
         if(!shmsize)
           fprintf(stderr, "Failed to parse memory size, using default.\n");
       }
+      if(argc>(i+2)) {
+        mode = parse_memmode(argv[i+2]);
+        if(mode == 0)
+          fprintf(stderr, "Invalid permission mode, using default.\n");
+      }
       shmptr=wg_attach_memsegment(shmname, shmsize, shmsize, 1,
-        (flags & FLAGS_LOGGING));
+        (flags & FLAGS_LOGGING), mode);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -586,30 +592,6 @@ int main(int argc, char **argv) {
       break;
     }
 #endif
-    else if(argc>(i+1) && !strcmp(argv[i], "fill")) {
-      int rows = atol(argv[i+1]);
-      if(!rows) {
-        fprintf(stderr, "Invalid number of rows.\n");
-        exit(1);
-      }
-
-      shmptr=wg_attach_database(shmname, shmsize);
-      if(!shmptr) {
-        fprintf(stderr, "Failed to attach to database.\n");
-        exit(1);
-      }
-
-      WLOCK(shmptr, wlock);
-      if(argc > (i+2) && !strcmp(argv[i+2], "mix"))
-        wg_genintdata_mix(shmptr, rows, TESTREC_SIZE);
-      else if(argc > (i+2) && !strcmp(argv[i+2], "desc"))
-        wg_genintdata_desc(shmptr, rows, TESTREC_SIZE);
-      else
-        wg_genintdata_asc(shmptr, rows, TESTREC_SIZE);
-      WULOCK(shmptr, wlock);
-      printf("Data inserted\n");
-      break;
-    }
     else if(argc>(i+1) && !strcmp(argv[i],"select")) {
       int rows = atol(argv[i+1]);
       int from = 0;
@@ -621,14 +603,14 @@ int main(int argc, char **argv) {
       if(argc > (i+2))
         from = atol(argv[i+2]);
 
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
       }
-      RLOCK(shmptr, wlock);
+      RLOCK(shmptr, rlock);
       selectdata(shmptr, rows, from);
-      RULOCK(shmptr, wlock);
+      RULOCK(shmptr, rlock);
       break;
     }
     else if(argc>(i+1) && !strcmp(argv[i],"add")) {
@@ -646,7 +628,7 @@ int main(int argc, char **argv) {
       break;
     }
     else if(argc>(i+2) && !strcmp(argv[i],"del")) {
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -657,7 +639,7 @@ int main(int argc, char **argv) {
       break;
     }
     else if(argc>(i+3) && !strcmp(argv[i],"query")) {
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -689,7 +671,7 @@ int main(int argc, char **argv) {
       break;
     }
     else if(argc>(i+1) && !strcmp(argv[i],"findjson")) {
-      shmptr=wg_attach_database(shmname, shmsize);
+      shmptr=wg_attach_existing_database(shmname);
       if(!shmptr) {
         fprintf(stderr, "Failed to attach to database.\n");
         exit(1);
@@ -699,6 +681,65 @@ int main(int argc, char **argv) {
       WULOCK(shmptr, wlock);
       break;
     }
+    else if(argc>(i+1) && !strcmp(argv[i], "createindex")) {
+      int col;
+      shmptr = (void *) wg_attach_database(shmname, shmsize);
+      if(!shmptr) {
+        fprintf(stderr, "Failed to attach to database.\n");
+        exit(1);
+      }
+      sscanf(argv[i+1], "%d", &col);
+      WLOCK(shmptr, wlock);
+      wg_create_index(shmptr, col, WG_INDEX_TYPE_TTREE, NULL, 0);
+      WULOCK(shmptr, wlock);
+      break;
+    }
+    else if(argc>(i+1) && !strcmp(argv[i], "createhash")) {
+      gint cols[MAX_INDEX_FIELDS], col_count, j;
+      shmptr = (void *) wg_attach_database(shmname, shmsize);
+      if(!shmptr) {
+        fprintf(stderr, "Failed to attach to database.\n");
+        exit(1);
+      }
+      col_count = 0;
+      for(j = i+1; j<argc; j++) {
+        int col;
+        sscanf(argv[j], "%d", &col);
+        cols[col_count++] = col;
+      }
+      WLOCK(shmptr, wlock);
+      wg_create_multi_index(shmptr, cols, col_count,
+        WG_INDEX_TYPE_HASH_JSON, NULL, 0);
+      WULOCK(shmptr, wlock);
+      break;
+    }
+    else if(argc>(i+1) && !strcmp(argv[i], "dropindex")) {
+      int index_id;
+      shmptr = (void *) wg_attach_database(shmname, shmsize);
+      if(!shmptr) {
+        fprintf(stderr, "Failed to attach to database.\n");
+        exit(1);
+      }
+      sscanf(argv[i+1], "%d", &index_id);
+      WLOCK(shmptr, wlock);
+      if(wg_drop_index(shmptr, index_id))
+        fprintf(stderr, "Failed to drop index.\n");
+      else
+        printf("Index dropped.\n");
+      WULOCK(shmptr, wlock);
+      break;
+    }
+    else if(!strcmp(argv[i], "listindex")) {
+      shmptr = (void *) wg_attach_database(shmname, shmsize);
+      if(!shmptr) {
+        fprintf(stderr, "Failed to attach to database.\n");
+        exit(1);
+      }
+      RLOCK(shmptr, rlock);
+      print_indexes(shmptr, stdout);
+      RULOCK(shmptr, rlock);
+      break;
+    }
 
     shmname = argv[1]; /* no match, assume shmname was given */
   }
@@ -929,7 +970,7 @@ wg_json_query_arg *make_json_arglist(void *db, char *json, int *sz,
   void *document;
   gint i, reclen;
 
-  if(wg_parse_json_param(db, json, &document))
+  if(wg_check_json(db, json) || wg_parse_json_param(db, json, &document))
     return NULL;
 
   if(!is_schema_object(document)) {
@@ -982,7 +1023,7 @@ void findjson(void *db, char *json) {
 
   rec = wg_fetch(db, q);
   while(rec) {
-    wg_print_json_document(db, stdout, rec);
+    wg_print_json_document(db, NULL, NULL, rec);
     printf("\n");
     rec = wg_fetch(db, q);
   }
@@ -995,6 +1036,115 @@ abort:
   }
 }
 
+/** Print information about the memory database.
+ */
+void segment_stats(void *db) {
+  char buf1[200], buf2[40];
+#ifndef _WIN32
+  struct passwd *pwd;
+  struct group *grp;
+#endif
+  db_memsegment_header *dbh = dbmemsegh(db);
+
+  printf("database key: %d\n", (int) dbh->key);
+  printf("database version: ");
+  wg_print_header_version(dbh, 0);
+  wg_pretty_print_memsize(dbh->size, buf1, 40);
+  wg_pretty_print_memsize(dbh->size - dbh->free, buf2, 40);
+  printf("free space: %s (of %s)\n", buf2, buf1);
+#ifndef _WIN32
+  pwd = getpwuid(wg_memowner(db));
+  if(pwd) {
+    printf("owner: %s\n", pwd->pw_name);
+  }
+  grp = getgrgid(wg_memgroup(db));
+  if(grp) {
+    printf("group: %s\n", grp->gr_name);
+  }
+  printf("permissions: %o\n", wg_memmode(db));
+#endif
+#ifdef USE_DBLOG
+  if(dbh->logging.active) {
+    wg_journal_filename(db, buf1, 200);
+    printf("logging is active, journal file: %s\n", buf1);
+  } else {
+    printf("logging is not active\n");
+  }
+#endif
+  printf("database has ");
+  switch(dbh->index_control_area_header.number_of_indexes) {
+    case 0:
+      printf("no indexes\n");
+      break;
+    case 1:
+      printf("1 index\n");
+      break;
+    default:
+      printf("%d indexes\n",
+        (int) dbh->index_control_area_header.number_of_indexes);
+      break;
+  }
+}
+
+void print_indexes(void *db, FILE *f) {
+  int column;
+  db_memsegment_header* dbh = dbmemsegh(db);
+  gint *ilist;
+
+  if(!dbh->index_control_area_header.number_of_indexes) {
+    fprintf(f, "No indexes in the database.\n");
+    return;
+  }
+  else {
+    fprintf(f, "col\ttype\tmulti\tid\tmask\n");
+  }
+
+  for(column=0; column<=MAX_INDEXED_FIELDNR; column++) {
+    ilist = &dbh->index_control_area_header.index_table[column];
+    while(*ilist) {
+      gcell *ilistelem = (gcell *) offsettoptr(db, *ilist);
+      if(ilistelem->car) {
+        char typestr[3];
+        wg_index_header *hdr = \
+          (wg_index_header *) offsettoptr(db, ilistelem->car);
+        typestr[2] = '\0';
+        switch(hdr->type) {
+          case WG_INDEX_TYPE_TTREE:
+            typestr[0] = 'T';
+            typestr[1] = '\0';
+            break;
+          case WG_INDEX_TYPE_TTREE_JSON:
+            typestr[0] = 'T';
+            typestr[1] = 'J';
+            break;
+          case WG_INDEX_TYPE_HASH:
+            typestr[0] = '#';
+            typestr[1] = '\0';
+            break;
+          case WG_INDEX_TYPE_HASH_JSON:
+            typestr[0] = '#';
+            typestr[1] = 'J';
+            break;
+          default:
+            break;
+        }
+        fprintf(f, "%d\t%s\t%d\t%d\t%s\n",
+          column,
+          typestr,
+          (int) hdr->fields,
+          (int) ilistelem->car,
+#ifndef USE_INDEX_TEMPLATE
+          "-");
+#else
+          (hdr->template_offset ? "Y" : "N"));
+#endif
+      }
+      ilist = &ilistelem->cdr;
+    }
+  }
+}
+
+
 #ifdef __cplusplus
 }
 #endif
